@@ -9,17 +9,15 @@
 #include "colorpic.c"
 #include "monopic.c"
 
-Pixmap bars_pixmap = None;
-Pixmap lock_pixmap = None;
-
 Gif_Stream *current_slideshow = 0;
 
 #define NPICTURES 14
 
 struct named_record {
   const char *name;
-  Gif_Record *record;
+  const Gif_Record *record;
   Gif_Stream *gfs;
+  const char *synonym;
 };
 
 struct named_record built_in_pictures[] = {
@@ -43,6 +41,9 @@ struct named_record built_in_pictures[] = {
 
   /* other cultures' finger gestures */
   {"korean", &koreanl_gif, 0},		{"koreanicon", &koreani_gif, 0},
+  {"german", &germanl_gif, 0},		{"germanicon", &germani_gif, 0},
+  {"japanese", 0, 0, "korean"},		{"japaneseicon", 0, 0, "koreanicon"},
+  {"russian", 0, 0, "korean"},		{"russianicon", 0, 0, "koreanicon"},
 
   /* last */
   {0, 0, 0}
@@ -53,33 +54,39 @@ static void
 free_picture(void *v)
 {
   Picture *p = (Picture *)v;
+  int i;
   if (p->canonical)
     Gif_DeleteImage(p->canonical);
-  else if (p->pix)
-    XFreePixmap(display, p->pix);
+  else
+    for (i = 0; i < nports; i++)
+      if (p->pix[i])
+	XFreePixmap(ports[i].display, p->pix[i]);
   xfree(p);
 }
 
 static void
 add_picture(Gif_Image *gfi, int clock_x_off, int clock_y_off)
 {
-  Picture *p = (Picture *)xmalloc(sizeof(Picture));
-  p->pix = None;
+  Picture *p = (Picture *)
+    xmalloc(sizeof(Picture) + (nports-1)*sizeof(Pixmap));
+  int i;
   p->clock_x_off = clock_x_off;
   p->clock_y_off = clock_y_off;
   p->canonical = 0;
+  for (i = 0; i < nports; i++)
+    p->pix[i] = None;
   
   gfi->user_data = p;
   gfi->free_user_data = free_picture;
 }
 
 
-Gif_Image *
+Gif_Stream *
 get_built_in_image(const char *name)
 {
   struct named_record *nr;
   Gif_Stream *gfs;
-  int xoff;
+  int i, xoff;
   
   for (nr = built_in_pictures; nr->name; nr++)
     if (strcmp(nr->name, name) == 0)
@@ -88,36 +95,132 @@ get_built_in_image(const char *name)
   
  found:
   if (nr->gfs)
-    return nr->gfs->images[0];
+    return nr->gfs;
+  if (!nr->record && nr->synonym) {
+    nr->gfs = get_built_in_image(nr->synonym);
+    return nr->gfs;
+  }
   
-  nr->gfs = gfs = Gif_FullReadRecord(nr->record, GIF_READ_COMPRESSED, 0, 0);
+  nr->gfs = gfs =
+    Gif_FullReadRecord(nr->record, GIF_READ_COMPRESSED | GIF_READ_CONST_RECORD,
+		       0, 0);
   if (!gfs)
     return 0;
   
   xoff = (strncmp(name, "locked", 6) == 0 ? 65 : 10);
-  add_picture(gfs->images[0], xoff, 10);
-  if (!gfs->images[0]->local) {
-    gfs->images[0]->local = gfs->global;
-    gfs->global->refcount++;
-  }
-  gfs->images[0]->delay = 0;
-  return gfs->images[0];
+  for (i = 0; i < gfs->nimages; i++)
+    add_picture(gfs->images[i], xoff, 10);
+
+  /* built-in images are all loop-forever. don't change the GIFs because it
+     makes the executable bigger */
+  if (gfs->loopcount < 0) gfs->loopcount = 0;
+  
+  return gfs;
 }
 
 
 #define MIN_DELAY 4
 
+static Gif_Image *
+clone_image_skeleton(Gif_Image *gfi)
+{
+  Gif_Image *ngfi = Gif_NewImage();
+  Picture *p = (Picture *)gfi->user_data;
+  assert(!gfi->image_data && gfi->compressed && p);
+  ngfi->local = gfi->local;
+  if (ngfi->local) ngfi->local->refcount++;
+  ngfi->transparent = gfi->transparent;
+  ngfi->left = gfi->left;
+  ngfi->top = gfi->top;
+  ngfi->width = gfi->width;
+  ngfi->height = gfi->height;
+  ngfi->compressed = gfi->compressed;
+  ngfi->compressed_len = gfi->compressed_len;
+  ngfi->free_compressed = 0;
+  ngfi->delay = gfi->delay;
+  add_picture(ngfi, p->clock_x_off, p->clock_y_off);
+  ((Picture *)ngfi->user_data)->canonical = gfi;
+  gfi->refcount++;
+  return ngfi;
+}
+
+static void
+add_stream_to_slideshow(Gif_Stream *add, Gif_Stream *gfs,
+			double flash_rate_ratio)
+{
+  Gif_Image *gfi;
+  int i;
+  double d;
+  
+  /* adapt delays for 1-frame images */
+  if (add->nimages == 1)
+    add->images[0]->delay = DEFAULT_FLASH_DELAY_SEC * 100;
+
+  /* account for background and loopcount */
+  if (gfs->nimages == 0) {
+    if (add->global) {
+      gfs->global = add->global;
+      add->global->refcount++;
+      gfs->background = add->background;
+    }
+    gfs->loopcount = add->loopcount;
+  }
+  
+  /* adapt screen size */
+  if (add->screen_width > gfs->screen_width)
+    gfs->screen_width = add->screen_width;
+  if (add->screen_height > gfs->screen_height)
+    gfs->screen_height = add->screen_height;
+  
+  /* add images from add to gfs */
+  for (i = 0; i < add->nimages; i++) {
+    gfi = add->images[i];
+    /* don't use an image directly if it is shared */
+    if (gfi->user_data)
+      gfi = clone_image_skeleton(gfi);
+    else
+      add_picture(gfi, 10, 10);
+    /* ensure local colormap */
+    if (!gfi->local && add->global) {
+      gfi->local = add->global;
+      add->global->refcount++;
+    }
+    /* adapt delay */
+    d = gfi->delay * flash_rate_ratio;
+    if (d < 0 || d >= 0xFFFF)
+      gfi->delay = 0xFFFF;
+    else
+      gfi->delay = (d < MIN_DELAY ? MIN_DELAY : (u_int16_t)d);
+    /* add image */
+    Gif_AddImage(gfs, gfi);
+  }
+  
+  /* add multiple times if it has a loop count, up to a max of 20 loops */
+  if (add->nimages > 1 && add->loopcount >= 0) {
+    int loop = (add->loopcount <= 20 ? add->loopcount : 20);
+    int first = gfs->nimages - add->nimages;
+    int j;
+    if (loop == 0 && first > 0) {
+      for (i = j = 0; i < add->nimages; i++)
+	j += gfs->images[first + i]->delay;
+      loop =
+	(int)((DEFAULT_FLASH_DELAY_SEC * 100) * flash_rate_ratio / j);
+    }
+    for (i = 0; i < loop; i++)
+      for (j = 0; j < add->nimages; j++)
+	Gif_AddImage(gfs, gfs->images[first + j]);
+  }
+}
+
 Gif_Stream *
 parse_slideshow(const char *slideshowtext, double flash_rate_ratio, int mono)
 {
   char buf[BUFSIZ];
   char name[BUFSIZ + 4];
   char *s;
-  Gif_Stream *gfs;
+  Gif_Stream *gfs, *add;
   Gif_Image *gfi;
-  u_int16_t delay;
   int i;
-  double d;
   
   if (strlen(slideshowtext) >= BUFSIZ) return 0;
   strcpy(buf, slideshowtext);
@@ -126,12 +229,6 @@ parse_slideshow(const char *slideshowtext, double flash_rate_ratio, int mono)
   gfs = Gif_NewStream();
   gfs->loopcount = 0;
   
-  d = flash_rate_ratio * DEFAULT_FLASH_DELAY_SEC * 100;
-  if (d < 0 || d > 0xFFFF)
-    delay = 0xFFFF;
-  else
-    delay = (d < MIN_DELAY ? MIN_DELAY : (u_int16_t)d);
-  
   while (*s) {
     char *n, save;
     
@@ -146,92 +243,31 @@ parse_slideshow(const char *slideshowtext, double flash_rate_ratio, int mono)
       strcpy(name, n + 1);
       if (mono) strcat(name, "mono");
       i = strlen(name);
-      gfi = get_built_in_image(name);
+      add = get_built_in_image(name);
       /* some images don't have monochromatic versions; fall back on color */
-      if (!gfi && i > 4 && strcmp(name + i - 4, "mono") == 0) {
+      if (!add && i > 4 && strcmp(name + i - 4, "mono") == 0) {
 	name[i-4] = 0;
-	gfi = get_built_in_image(name);
-	if (gfi)
+	add = get_built_in_image(name);
+	if (add)
 	  warning("no monochrome version of built-in picture `%s'", name);
 	name[i-4] = 'm';
       }
-      if (!gfi)
+      if (!add)
 	error("unknown built-in picture `%s'", name);
-      /* adapt delay */
-      if (!gfi->delay || gfi->delay == delay)
-	gfi->delay = delay;
-      else {
-	Gif_Image *ngfi = Gif_NewImage();
-	Picture *p = (Picture *)gfi->user_data;
-	assert(!gfi->image_data && gfi->compressed);
-	ngfi->local = gfi->local;
-	ngfi->local->refcount++;
-	ngfi->transparent = gfi->transparent;
-	ngfi->left = gfi->left;
-	ngfi->top = gfi->top;
-	ngfi->width = gfi->width;
-	ngfi->height = gfi->height;
-	ngfi->compressed = gfi->compressed;
-	ngfi->compressed_len = gfi->compressed_len;
-	ngfi->free_compressed = 0;
-	ngfi->delay = delay;
-	add_picture(ngfi, p->clock_x_off, p->clock_y_off);
-	((Picture *)ngfi->user_data)->canonical = gfi;
-	gfi->refcount++;
-	gfi = ngfi;
-      }
-      /* add image */
-      Gif_AddImage(gfs, gfi);
+      /* add images */
+      add_stream_to_slideshow(add, gfs, flash_rate_ratio);
       
     } else {
       /* load file from disk */
       FILE *f = fopen(n, "rb");
-      Gif_Stream *read = Gif_FullReadFile(f, GIF_READ_COMPRESSED, 0, 0);
+      add = Gif_FullReadFile(f, GIF_READ_COMPRESSED, 0, 0);
       if (!f)
 	error("%s: %s", n, strerror(errno));
-      else if (!read || (read->nimages == 0 && read->errors > 0))
+      else if (!add || (add->nimages == 0 && add->errors > 0))
 	error("%s: not a GIF", n);
-      else {
-	/* adapt delays */
-	if (read->nimages == 1)
-	  read->images[0]->delay = delay;
-	else
-	  for (i = 0; i < read->nimages; i++) {
-	    gfi = read->images[i];
-	    d = gfi->delay * flash_rate_ratio;
-	    if (d < 0 || d >= 0xFFFF)
-	      gfi->delay = 0xFFFF;
-	    else
-	      gfi->delay = (d < MIN_DELAY ? MIN_DELAY : (u_int16_t)d);
-	  }
-	/* account for background and loopcount */
-	if (gfs->nimages == 0) {
-	  if (read->global) {
-	    gfs->global = read->global;
-	    read->global->refcount++;
-	    gfs->background = read->background;
-	  }
-	  gfs->loopcount = read->loopcount;
-	}
-	/* adapt screen size */
-	if (read->screen_width > gfs->screen_width)
-	  gfs->screen_width = read->screen_width;
-	if (read->screen_height > gfs->screen_height)
-	  gfs->screen_height = read->screen_height;
-	/* add images to gfs */
-	for (i = 0; i < read->nimages; i++) {
-	  Gif_Image *gfi = read->images[i];
-	  if (!gfi->local && read->global) {
-	    gfi->local = read->global;
-	    read->global->refcount++;
-	  }
-	  Gif_AddImage(gfs, gfi);
-	  add_picture(gfi, 10, 10);
-	  if (gfi->delay == 0)
-	    gfi->delay = 1;
-	}
-      }
-      if (read) Gif_DeleteStream(read);
+      else
+	add_stream_to_slideshow(add, gfs, flash_rate_ratio);
+      if (add) Gif_DeleteStream(add);
       if (f) fclose(f);
     }
     
@@ -256,39 +292,48 @@ parse_slideshow(const char *slideshowtext, double flash_rate_ratio, int mono)
 
 
 void
-set_slideshow(Hand *h, Gif_Stream *gfs, struct timeval *now)
+set_slideshow(Hand *h, Gif_Stream *gfs, const struct timeval *now_ptr)
 {
-  int i = 0;
+  int which_im = 0;
   Alarm *a;
   struct timeval t;
+  Port *port = h->port;
   
-  assert(gfs);
   if (h->slideshow == gfs)
     return;
+  else if (!gfs) {
+    h->slideshow = 0;
+    unschedule_data(A_FLASH, h, 0);
+    return;
+  }
   
-  if (now)
-    t = *now;
+  if (now_ptr)
+    t = *now_ptr;
   else
     xwGETTIME(t);
   
-  a = grab_alarm_data(A_FLASH, h);
+  a = grab_alarm_data(A_FLASH, h, 0);
   
   if (h->slideshow) {
     Gif_Image *cur_im = h->slideshow->images[h->slide];
-    if (a) xwSUBDELAY(t, a->timer, cur_im->delay);
-    /* Leave i with a picture that matches the old one, or 0 if no
+    if (a)
+      xwSUBDELAY(t, a->timer, cur_im->delay);
+    /* Leave which_im with a picture that matches the old one, or 0 if no
        picture matches the old one. */
-    for (i = gfs->nimages - 1; i > 0; i--)
-      if (gfs->images[i] == cur_im)
+    for (which_im = gfs->nimages - 1; which_im > 0; which_im--)
+      if (gfs->images[which_im] == cur_im)
 	break;
   }
   
+  /* fprintf(stderr, "%p %p %d\n", h, gfs, which_im); */
   if (gfs->nimages > 1) {
-    if (!a) a = new_alarm_data(A_FLASH, h);
-    xwADDDELAY(a->timer, t, gfs->images[i]->delay);
+    if (!a)
+      a = new_alarm_data(A_FLASH, h, 0);
+    xwADDDELAY(a->timer, t, gfs->images[which_im]->delay);
     schedule(a);
   } else {
-    if (a) destroy_alarm(a);
+    if (a)
+      destroy_alarm(a);
   }
   
   h->slideshow = gfs;
@@ -302,14 +347,14 @@ set_slideshow(Hand *h, Gif_Stream *gfs, struct timeval *now)
     xsh->min_height = xsh->max_height = gfs->screen_height;
     wmch.width = gfs->screen_width;
     wmch.height = gfs->screen_height;
-    XSetWMNormalHints(display, h->w, xsh);
-    XReconfigureWMWindow(display, h->w, port.screen_number,
+    XSetWMNormalHints(port->display, h->w, xsh);
+    XReconfigureWMWindow(port->display, h->w, port->screen_number,
 			 CWWidth | CWHeight, &wmch);
     XFree(xsh);
   }
   
   h->loopcount = 0;
-  h->slide = i;
+  h->slide = which_im;
   draw_slide(h);
 }
 
